/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry.jdbc;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.Collection;
import org.wso2.registry.*;
import org.wso2.registry.config.RegistryContext;
import org.wso2.registry.jdbc.dao.*;
import org.wso2.registry.jdbc.handlers.HandlerManager;
import org.wso2.registry.jdbc.handlers.Handler;
import org.wso2.registry.jdbc.handlers.filters.MediaTypeMatcher;
import org.wso2.registry.jdbc.handlers.filters.URLMatcher;
import org.wso2.registry.jdbc.handlers.filters.Filter;
import org.wso2.registry.jdbc.handlers.builtin.*;
import org.wso2.registry.jdbc.queries.QueryProcessorManager;
import org.wso2.registry.jdbc.realm.RegistryRealm;
import org.wso2.registry.jdbc.utils.Transaction;
import org.wso2.registry.session.CurrentSession;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.users.UserStoreException;
import org.wso2.registry.users.accesscontrol.AuthorizingRealm;
import org.wso2.registry.users.accesscontrol.AuthorizingRealmConfig;
import org.wso2.registry.utils.AuthorizationUtils;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Map;
import java.util.Date;
import java.util.HashMap;

/**
 * JDBC based implementation of the Registry. This will be used mostly as the back-end by other
 * Registry implementations. This can use either an in-memory database or a external database
 * configured using data source.
 */
public class JDBCRegistry implements Registry {

    private TransactionalJDBCRegistry transactionalRegistry;
    private NonTransactionalJDBCRegistry nonTransactionalRegistry;

    private static final Log log = LogFactory.getLog(JDBCRegistry.class);

    private DataSource dataSource = null;
    private ResourceDAO resourceDAO = null;

    protected HandlerManager handlerManager;
    protected Repository repository;
    protected VersionRepository versionRepository;
    private Map<String, Aspect> aspects = new HashMap<String, Aspect>();

    private UserRealm defaultRealm = null;
    protected RegistryContext registryContext;

    private UserRealm systemUserRealm;

    /**
     * Default constructor. JDBC registry should be configured using the configure(...) method, if
     * it is instantiated using this constructor.
     */
    public JDBCRegistry() {
    }

    /**
     * Constructs a JDBC registry to be used without any security (authorizations on actions). JDBC
     * registry instances created using this constructor should not be used with SecureRegistry
     * instances.
     *
     * @param dataSource Data source to be used to store resources and metadata.
     * @throws RegistryException : If something went wrong while init
     */
    public JDBCRegistry(DataSource dataSource) throws RegistryException {
        this.dataSource = dataSource;
        init(null);
    }

    /**
     * Constructs a JDBC registry to be used with secure registries. Default authorizations will be
     * applied on the resources created using this registry.
     *
     * @param dataSource Data source to be used to store resources and metadata. Note that the same
     *                   data source may or may not be used as the data source of user manager.
     * @param realm      User manager realm to handle authorizations. It is strongly recommended to
     *                   use a realm instance constructed using the RegistryRealmFactory. All
     *                   initial users, roles and authorizations will be set on realms obtained
     *                   using the RegistryRealmFactory.
     * @throws RegistryException : If something went wrong
     */
    public JDBCRegistry(DataSource dataSource, UserRealm realm) throws RegistryException {
        this.dataSource = dataSource;
        init(realm);
    }

    /**
     * Configures and initiates the JDBC registry with a (new) datasource and a realm. This is
     * useful for changing underlying databases at run-time.
     *
     * @param dataSource the DataSource to use
     * @param realm      Realm to use
     * @throws RegistryException : If something went wrong while init
     */
    public void configure(DataSource dataSource, RegistryRealm realm) throws RegistryException {
        this.dataSource = dataSource;
        init(realm);
    }

    private void init(UserRealm realm) throws RegistryException {

        defaultRealm = realm;

        try {
            systemUserRealm = new AuthorizingRealm();
            AuthorizingRealmConfig config =
                    (AuthorizingRealmConfig) systemUserRealm.getRealmConfiguration();
            config.setRealm(defaultRealm);
            config.setAuthenticatedUserName(RegistryConstants.SYSTEM_USER);
            config.setCurrentUserReadable(true);
            config.setCurrentUserEditable(true);
            systemUserRealm.init(config);

        } catch (UserStoreException e) {
            String msg = "Failed to initialize the system realm. " + e.getMessage();
            log.fatal(msg, e);
            throw new RegistryException(msg, e);
        }

        repository = new Repository(dataSource);
        versionRepository = new VersionRepository(dataSource);
        handlerManager = new HandlerManager(this, repository);
        QueryProcessorManager queryProcessorManager =
                new QueryProcessorManager(dataSource, realm, this);

        RegistryContext registryContext = RegistryContext.getRegistryContext();
        //RegistryContext registryContext = (RegistryContext)System.
        //        getProperties().get(RegistryConstants.REGISTRY_CONTEXT);

        if (registryContext == null) {
            registryContext = new RegistryContext();
            RegistryContext.setRegistryContext(registryContext);
        }
        registryContext.setRepository(repository);

        resourceDAO = new ResourceDAO();

        // check if root is added. if not add it.
        addRootCollection();

        registerBuiltInHandlers();

        transactionalRegistry =
                new TransactionalJDBCRegistry(registryContext, handlerManager, repository,
                                              versionRepository, queryProcessorManager, aspects);

        nonTransactionalRegistry = new NonTransactionalJDBCRegistry(
                transactionalRegistry, dataSource);
    }

    private void registerBuiltInHandlers() {

        Axis2RepositoryHandler axis2RepositoryHandler = new Axis2RepositoryHandler();
        MediaTypeMatcher axis2MediaTypeMatcher =
                new MediaTypeMatcher("application/vnd.apache.axis2");
        handlerManager.addHandler(0, axis2MediaTypeMatcher, axis2RepositoryHandler);

        SQLQueryHandler sqlQueryHandler = new SQLQueryHandler();
        MediaTypeMatcher sqlMediaTypeMatcher =
                new MediaTypeMatcher(RegistryConstants.SQL_QUERY_MEDIA_TYPE);
        handlerManager.addHandler(0, sqlMediaTypeMatcher, sqlQueryHandler);

        CommentURLHandler commentURLHandler = new CommentURLHandler();
        URLMatcher commentURLMatcher = new URLMatcher();
        commentURLMatcher.setGetPattern(".+;comments:[0-9]+");
        commentURLMatcher.setDeletePattern(".+;comments:[0-9]+");
        handlerManager.addHandler(0, commentURLMatcher, commentURLHandler);

        CommentCollectionURLHandler commentCollectionURLHandler = new CommentCollectionURLHandler();
        URLMatcher commentCollectionURLMatcher = new URLMatcher();
        commentCollectionURLMatcher.setGetPattern(".+;comments");
        handlerManager.addHandler(0, commentCollectionURLMatcher, commentCollectionURLHandler);

        RatingURLHandler ratingURLHandler = new RatingURLHandler();
        URLMatcher ratingURLMatcher = new URLMatcher();
        ratingURLMatcher.setGetPattern(".+;ratings:.+");
        handlerManager.addHandler(0, ratingURLMatcher, ratingURLHandler);

        RatingCollectionURLHandler ratingCollectionURLHandler = new RatingCollectionURLHandler();
        URLMatcher ratingCollectionURLMatcher = new URLMatcher();
        ratingCollectionURLMatcher.setGetPattern(".+;ratings");
        handlerManager.addHandler(0, ratingCollectionURLMatcher, ratingCollectionURLHandler);

        TagURLHandler tagURLHandler = new TagURLHandler();
        URLMatcher tagURLMatcher = new URLMatcher();
        tagURLMatcher.setGetPattern(".+;.+:.+:.+");
        handlerManager.addHandler(0, tagURLMatcher, tagURLHandler);

        //TagCollectionURLHandler tagCollectionURLHandler = new TagCollectionURLHandler();
        //URLMatcher tagURLMatcher = new URLMatcher();
        //tagURLMatcher.setGetPattern(".+;tags");
        //handlerManager.addHandler(0, tagURLMatcher, tagCollectionURLHandler);
        //
        //TagURLHandler tagURLHandler = new TagURLHandler();
        //tagURLMatcher = new URLMatcher();
        //tagURLMatcher.setGetPattern(".+;tags:.+:.+");
        //handlerManager.addHandler(0, tagURLMatcher, tagCollectionURLHandler);
    }

    public void addHandler(Filter filter, Handler handler) {
        handlerManager.addHandler(0, filter, handler);
    }

    public void beginTransaction() throws RegistryException {

        if (Transaction.isStarted()) {
            String msg = "Failed to start new transaction. Transaction is already " +
                    "started by this thread. Please close started transactions before " +
                    "starting new transaction.";

            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn;
        try {
            conn = dataSource.getConnection();
            conn.setAutoCommit(false);
        } catch (SQLException e) {
            String msg = "Failed to start new registry transaction. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        Transaction.setConnection(conn);
    }

    public void rollbackTransaction() throws RegistryException {

        if (!Transaction.isStarted()) {
            String msg = "Failed to rollback transaction. A transaction is not started.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn = Transaction.getConnection();
        try {
            conn.rollback();

        } catch (SQLException e) {
            String msg = "Failed to rollback transaction. " + e.getMessage();
            log.error(msg);
            throw new RegistryException(msg);

        } finally {
            endTransaction();
        }
    }

    public void commitTransaction() throws RegistryException {

        if (!Transaction.isStarted()) {
            String msg = "Failed to commit transaction. A transaction is not started.";
            log.error(msg);
            throw new RegistryException(msg);
        }

        Connection conn = Transaction.getConnection();
        try {
            conn.commit();

        } catch (SQLException e) {
            String msg = "Failed to commit transaction. " + e.getMessage();
            log.error(msg);
            throw new RegistryException(msg);

        } finally {
            endTransaction();
        }
    }

    private void endTransaction() throws RegistryException {

        if (Transaction.isStarted()) {
            Connection conn = Transaction.getConnection();
            try {
                conn.close();

            } catch (SQLException e) {
                String msg = "Failed to close transaction. " + e.getMessage();
                log.error(msg);
                throw new RegistryException(msg);

            } finally {
                Transaction.setStarted(false);
                Transaction.setConnection(null);
            }
        }
    }

    public Resource newResource() throws RegistryException {
        ResourceImpl resource = new ResourceImpl();
        resource.setAuthorUserName(CurrentSession.getUser());
        return resource;
    }

    public Collection newCollection() throws RegistryException {
        CollectionImpl coll = new CollectionImpl();
        coll.setAuthorUserName(CurrentSession.getUser());
        return coll;
    }

    public UserRealm getDefaultRealm() {
        return defaultRealm;
    }

    public Resource get(String path) throws RegistryException {

        Resource resource;
        if (Transaction.isStarted()) {
            resource = transactionalRegistry.get(path);
        } else {
            resource = nonTransactionalRegistry.get(path);
        }

        return resource;
    }

    public Collection get(String path, int start, int pageSize) throws RegistryException {
        Collection collection;
        if (Transaction.isStarted()) {
            collection = transactionalRegistry.get(path, start, pageSize);
        } else {
            collection = nonTransactionalRegistry.get(path, start, pageSize);            
        }
        return collection;
    }

    public boolean resourceExists(String path) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.resourceExists(path);
        } else {
            return nonTransactionalRegistry.resourceExists(path);
        }
    }

    public String put(String suggestedPath, Resource resource) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.put(suggestedPath, resource);
        } else {
            return nonTransactionalRegistry.put(suggestedPath, resource);
        }
    }

    public void delete(String path) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.delete(path);
        } else {
            nonTransactionalRegistry.delete(path);
        }
    }

    public String importResource(String suggestedPath, String sourceURL, Resource metadata) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.importResource(suggestedPath, sourceURL, metadata);
        } else {
            return nonTransactionalRegistry.importResource(suggestedPath, sourceURL, metadata);
        }
    }

    public String rename(String currentPath, String newPath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.rename(currentPath, newPath);
        } else {
            return nonTransactionalRegistry.rename(currentPath, newPath);
        }
    }

    public String move(String currentPath, String newPath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.move(currentPath, newPath);
        } else {
            return nonTransactionalRegistry.move(currentPath, newPath);
        }
    }

    public String copy(String sourcePath, String targetPath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.copy(sourcePath, targetPath);
        } else {
            return nonTransactionalRegistry.copy(sourcePath, targetPath);
        }
    }

    public void createVersion(String path) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.createVersion(path);
        } else {
            nonTransactionalRegistry.createVersion(path);
        }
    }

    public String[] getVersions(String path) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getVersions(path);
        } else {
            return nonTransactionalRegistry.getVersions(path);
        }
    }

    public void restoreVersion(String versionPath) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.restoreVersion(versionPath);
        } else {
            nonTransactionalRegistry.restoreVersion(versionPath);
        }
    }


    ////////////////////////////////////////////////////////
    // Associations
    ////////////////////////////////////////////////////////
    public void addAssociation(String associationType, String sourcePath, String associationPaths)
            throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.addAssociation(associationType, sourcePath, associationPaths);
        } else {
            nonTransactionalRegistry.addAssociation(associationType, sourcePath, associationPaths);
        }
    }

    public void removeAssociation(
            String associationType, String sourcePath, String associationPaths)
            throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.removeAssociation(associationType, sourcePath, associationPaths);
        } else {
            nonTransactionalRegistry.
                    removeAssociation(associationType, sourcePath, associationPaths);
        }
    }

    public Association[] getAllAssociations(String resourcePath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getAllAssociations(resourcePath);
        } else {
            return nonTransactionalRegistry.getAllAssociations(resourcePath);
        }
    }

    public Association[] getAssociations(String resourcePath, String associationType)
            throws RegistryException {

         if (Transaction.isStarted()) {
            return transactionalRegistry.getAssociations(resourcePath, associationType);
        } else {
            return nonTransactionalRegistry.getAssociations(resourcePath, associationType);
        }
    }

    public void applyTag(String resourcePath, String tag) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.applyTag(resourcePath, tag);
        } else {
            nonTransactionalRegistry.applyTag(resourcePath, tag);
        }
    }

    public TaggedResourcePath[] getResourcePathsWithTag(String tag) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getResourcePathsWithTag(tag);
        } else {
            return nonTransactionalRegistry.getResourcePathsWithTag(tag);
        }
    }

    public Tag[] getTags(String resourcePath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getTags(resourcePath);
        } else {
            return nonTransactionalRegistry.getTags(resourcePath);
        }
    }

    public void removeTag(String path, String tag) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.removeTag(path, tag);
        } else {
            nonTransactionalRegistry.removeTag(path, tag);
        }
    }

    public String addComment(String resourcePath, Comment comment) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.addComment(resourcePath, comment);
        } else {
            return nonTransactionalRegistry.addComment(resourcePath, comment);
        }
    }

    public void editComment(String commentPath, String text) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.editComment(commentPath, text);
        } else {
            nonTransactionalRegistry.editComment(commentPath, text);
        }
    }

    public Comment[] getComments(String resourcePath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getComments(resourcePath);
        } else {
            return nonTransactionalRegistry.getComments(resourcePath);
        }
    }

    public void rateResource(String resourcePath, int rating) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.rateResource(resourcePath, rating);
        } else {
            nonTransactionalRegistry.rateResource(resourcePath, rating);
        }
    }

    public float getAverageRating(String resourcePath) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getAverageRating(resourcePath);
        } else {
            return nonTransactionalRegistry.getAverageRating(resourcePath);
        }
    }

    public int getRating(String path, String userName) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getRating(path, userName);
        } else {
            return nonTransactionalRegistry.getRating(path, userName);
        }
    }

    public Collection executeQuery(String path, Map parameters) throws RegistryException {

        Collection resultCollection;
        if (Transaction.isStarted()) {
            resultCollection = transactionalRegistry.executeQuery(path, parameters);
        } else {
            resultCollection = nonTransactionalRegistry.executeQuery(path, parameters);
        }

        return resultCollection;
    }

    public LogEntry[] getLogs(String resourcePath, int action, String userName, Date from, Date to, boolean recentFirst) throws RegistryException {
        if (Transaction.isStarted()) {
            return transactionalRegistry.
                    getLogs(resourcePath, action, userName, from, to, recentFirst);
        } else {
            return nonTransactionalRegistry.
                    getLogs(resourcePath, action, userName, from, to, recentFirst);
        }
    }


    public LogEntryCollection getLogCollection(String resourcePath,
                                               int action,
                                               String userName,
                                               Date from,
                                               Date to,
                                               boolean recentFirst) throws RegistryException {
        LogEntryCollection logEntryCollection ;
        if (Transaction.isStarted()) {
            logEntryCollection = transactionalRegistry.
                    getLogCollection(resourcePath, action, userName, from, to, recentFirst);
        } else {
            logEntryCollection = nonTransactionalRegistry.
                    getLogCollection(resourcePath, action, userName, from, to, recentFirst);
        }
        logEntryCollection.setDataSource(dataSource);
        logEntryCollection.setResourcePath(resourcePath);
        logEntryCollection.setAction(action);
        logEntryCollection.setUserName(userName);
        logEntryCollection.setFrom(from);
        logEntryCollection.setTo(to);
        logEntryCollection.setRecentFirst(recentFirst);
        return logEntryCollection;

    }

    public void addAspect(String name, Aspect aspect) throws RegistryException {
        aspects.put(name, aspect);
    }

    public void associateAspect(String resourcePath, String aspect) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.associateAspect(resourcePath, aspect);
        } else {
            nonTransactionalRegistry.associateAspect(resourcePath, aspect);
        }
    }

    public void invokeAspect(String resourcePath, String aspectName, String action) throws RegistryException {

        if (Transaction.isStarted()) {
            transactionalRegistry.invokeAspect(resourcePath, aspectName, action);
        } else {
            nonTransactionalRegistry.invokeAspect(resourcePath, aspectName, action);
        }
    }

    public String[] getAspectActions(String resourcePath, String aspectName) throws RegistryException {

        if (Transaction.isStarted()) {
            return transactionalRegistry.getAspectActions(resourcePath, aspectName);
        } else {
            return nonTransactionalRegistry.getAspectActions(resourcePath, aspectName);
        }
    }

    private void addRootCollection() throws RegistryException {

        CurrentSession.setUser(RegistryConstants.SYSTEM_USER);
        CurrentSession.setRealm(systemUserRealm);

        try {

            beginTransaction();

            if (!resourceDAO.resourceExists(RegistryConstants.ROOT_PATH)) {

                CollectionImpl root = new CollectionImpl();
                resourceDAO.addRoot(RegistryConstants.ROOT_PATH, root);

                AuthorizationUtils.setRootAuthorizations(root.getId(), defaultRealm);
            }

            commitTransaction();

        } catch (Exception e) {

            String msg = "Failed to add the root collection to the registry.";
            log.fatal(msg, e);

            rollbackTransaction();

            throw new RegistryException(msg, e);
        }

        CurrentSession.removeUser();
        CurrentSession.removeRealm();

    }

    public void setRegistryContext(RegistryContext registryContext) {
        // TODO - ** HACK ALERT **!  This should all get initialized much more cleanly.
        if (registryContext.getDataSource() == null) {
            registryContext.setDataSource(dataSource);
        }

        this.registryContext = registryContext;
    }

    public RegistryContext getRegistryContext() {
        return registryContext;
    }
}
