/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.wso2.registry.jdbc.utils.Transaction;
import org.wso2.registry.jdbc.dataobjects.ResourceDO;
import org.wso2.registry.*;
import org.wso2.registry.Collection;
import org.wso2.registry.utils.AuthorizationUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.*;
import java.util.*;
import java.io.*;

public class ResourceVersionDAO {

    private static Log log = LogFactory.getLog(ResourceVersionDAO.class);

    public long addSnapshot(String snapshotRootID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO SNAPSHOT (ROOT_ID) VALUES (?)";
        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, snapshotRootID);
            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add snapshot for the resource " +
                    snapshotRootID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        String sql2 = "SELECT MAX(SNAPSHOT_ID) FROM SNAPSHOT WHERE ROOT_ID=? GROUP BY ROOT_ID";
        try {
            PreparedStatement ps2 = conn.prepareStatement(sql2);
            ps2.setString(1, snapshotRootID);

            ResultSet result = ps2.executeQuery();
            long snapshotID = 0;
            if (result.next()) {
                snapshotID = result.getLong(1);
            }

            return snapshotID;

        } catch (SQLException e) {

            String msg = "Failed to get the latest snapshot ID of the resource " +
                    snapshotRootID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public int getLatestResourceVersionNumber(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT MAX(VERSOIN) " +
                "FROM RESOURCE_VERSION WHERE RID=? GROUP BY RID";

        int snapshotNumber = 0;
        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                snapshotNumber = result.getInt(1);
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to get the latest snapshot number of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return snapshotNumber;
    }

    public int addResourceVersion(ResourceImpl resourceImpl, String contentVersionID)
            throws RegistryException {

        int versionNumber = getMaxResourceVersion(resourceImpl.getId()) + 1;

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO RESOURCE_VERSION (RID, VERSION, PATH, MEDIA_TYPE, COLLECTION, " +
                "CREATOR, CREATED_TIME, LAST_UPDATOR, LAST_UPDATED_TIME, DESCRIPTION, CONTENT_ID, " +
                "ASSOCIATED_SNAPSHOT_ID) " +
                "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            ps.setString(1, resourceImpl.getId());
            ps.setInt(2, versionNumber);
            ps.setString(3, resourceImpl.getPath());
            ps.setString(4, resourceImpl.getMediaType());
            ps.setInt(5, (resourceImpl instanceof Collection) ? 1: 0);
            ps.setString(6, resourceImpl.getAuthorUserName());
            ps.setTimestamp(7, new Timestamp(resourceImpl.getCreatedTime().getTime()));
            ps.setString(8, resourceImpl.getLastUpdaterUserName());
            ps.setTimestamp(9, new Timestamp(resourceImpl.getLastModified().getTime()));
            ps.setString(10, resourceImpl.getDescription());
            ps.setString(11, contentVersionID);
            ps.setLong(12, resourceImpl.getMatchingSnapshotID());

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add new version for resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return versionNumber;
    }

    public int getMaxResourceVersion(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT MAX(VERSION) FROM RESOURCE_VERSION WHERE RID=? GROUP BY RID";

        int maxVersion = 0;
        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                maxVersion = result.getInt(1);
            }

        } catch (SQLException e) {

            String msg = "Failed to latest version number of the resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return maxVersion;
    }

    public void addDependencies(String resourceID, int versionNumber, List<String> childIDs)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO DEPENDENCY_VERSION (PARENT_RID, PARENT_VERSION, CHILD_RID) " +
                "VALUES (?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);

            Iterator <String> iChildIDs = childIDs.iterator();
            while(iChildIDs.hasNext()) {
                String childID = iChildIDs.next();

                ps.setString(1, resourceID);
                ps.setInt(2, versionNumber);
                ps.setString(3, childID);

                ps.executeUpdate();
                ps.clearParameters();
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add dependencies for version " + versionNumber +
                    " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void addProperties(ResourceImpl resource, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "INSERT INTO PROPERTY_VERSION (RID, VERSION, NAME, VALUE) " +
                    "VALUES (?, ?, ?, ?)";

            PreparedStatement ps = conn.prepareStatement(sql);

            String resourceID = resource.getId();
            Properties props = resource.getProperties();
            if (props != null) {

                Iterator i = props.keySet().iterator();
                while (i.hasNext()) {
                    String name = (String)i.next();

                    List propValues = (List) props.get(name);
                    if (propValues != null) {
                        Iterator<String> iValues = propValues.iterator();
                        while (iValues.hasNext()) {
                            String value = iValues.next();

                            ps.setString(1, resourceID);
                            ps.setLong(2, version);
                            ps.setString(3, name);
                            ps.setString(4, value);

                            ps.executeUpdate();
                            ps.clearParameters();
                        }
                    }
                }
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to version properties to the resource " +
                    resource.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public void addSnapshotResource(long snapshotID, String resourceID, int versionNumber)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "INSERT INTO SNAPSHOT_RESOURCE_VERSION (SNAPSHOT_ID, RID, VERSION) " +
                "VALUES (?, ?, ?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);
            ps.setInt(3, versionNumber);

            ps.executeUpdate();
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add snapshot resource version for version " +
                    versionNumber + " of resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public Long[] getSnapshotIDs(String resourcePath) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT SNAPSHOT_ID FROM SNAPSHOT " +
                "WHERE ROOT_ID=(SELECT RID FROM RESOURCE WHERE PATH=?) ORDER BY SNAPSHOT_ID DESC";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourcePath);

            ResultSet results = ps.executeQuery();
            List <Long> snapshotIDs = new ArrayList();
            while(results.next()) {
                long snapshotNumber = results.getLong("SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }
            ps.close();

            return snapshotIDs.toArray(new Long[snapshotIDs.size()]);

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourcePath + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public long getVersion(String resourceID, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT VERSION FROM SNAPSHOT_RESOURCE_VERSION WHERE SNAPSHOT_ID=? AND RID=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setLong(1, snapshotID);
            ps.setString(2, resourceID);

            ResultSet result = ps.executeQuery();
            long version = -1;
            if (result.next()) {
                version = result.getLong("VERSION");
            }
            ps.close();

            return version;

        } catch (SQLException e) {

            String msg = "Failed to get version of resource " + resourceID +
                    " of snapshot " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public ResourceImpl get(String resourceID, long snapshotID, long version)
            throws RegistryException {

        ResourceImpl resourceImpl = getResource(resourceID, version);

        if (resourceImpl instanceof Collection) {
            fillChildren(resourceImpl, snapshotID);
        }

        fillResourceProperties(resourceImpl);

        return resourceImpl;
    }

    private void fillResourceProperties(ResourceImpl resourceImpl) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String propSQL = "SELECT NAME, VALUE FROM PROPERTY_VERSION WHERE RID=? AND VERSION=?";
            PreparedStatement ps = conn.prepareStatement(propSQL);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());
            ResultSet results = ps.executeQuery();

            while (results.next()) {
                String name = results.getString("NAME");
                String value = results.getString("VALUE");
                resourceImpl.addProperty(name, value);
            }

            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to add properties to the resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        resourceImpl.setPropertiesModified(false);
    }

    private void fillChildren(ResourceImpl resourceImpl, long snapshotID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            //String sql = "SELECT R.RID, R.PATH FROM RESOURCE_VERSION R, DEPENDENCY_VERSION D " +
            //        "WHERE D.PARENT_RID=? AND PARENT_VERSION=? AND D.CHILD_RID=R.RID AND R.VERSION=?";

            String sql = "SELECT CHILD_RID FROM DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceImpl.getId());
            ps.setLong(2, resourceImpl.getVersionNumber());

            List <String> childPathList = new ArrayList();
            ResultSet results = ps.executeQuery();
            while (results.next()) {

                String childID = results.getString("CHILD_RID");

                String sql2 = "SELECT R.PATH FROM RESOURCE_VERSION R " +
                        "WHERE R.RID=? AND R.VERSION=(" +
                        "SELECT S.VERSION FROM SNAPSHOT_RESOURCE_VERSION S " +
                        "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

                PreparedStatement ps2 = conn.prepareStatement(sql2);
                ps2.setString(1, childID);
                ps2.setLong(2, snapshotID);
                ps2.setString(3, childID);

                ResultSet result2 = ps2.executeQuery();
                String path = "";
                if (result2.next()) {
                    path = result2.getString("PATH");
                }
                ps2.close();

                if (AuthorizationUtils.authorize(childID, ActionConstants.GET)) {
                    String childPath = path + ";version=" + snapshotID;
                    childPathList.add(childPath);
                }
            }
            ps.close();

            String[] childPaths = childPathList.toArray(new String[childPathList.size()]);
            resourceImpl.setContent(childPaths);

        } catch (SQLException e) {

            String msg = "Failed to get child paths of resource " +
                    resourceImpl.getPath() + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    private ResourceImpl getResource(String resourceID, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        try {
            String sql = "SELECT RID, PATH, MEDIA_TYPE, COLLECTION, CREATOR, CREATED_TIME, " +
                    "LAST_UPDATOR, LAST_UPDATED_TIME, DESCRIPTION, CONTENT_ID " +
                    "FROM RESOURCE_VERSION WHERE RID=? AND VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            ResultSet result = ps.executeQuery();

            ResourceImpl resourceImpl = null;

            if (result.next()) {

                if (result.getInt("COLLECTION") == 0) {
                    resourceImpl = new ResourceVersionImpl();
                } else {
                    resourceImpl = new CollectionVersionImpl();
                }

                // this is always the current version of the resource
                resourceImpl.setVersionNumber(version);

                resourceImpl.setId(result.getString("RID"));
                resourceImpl.setPath(result.getString("PATH"));
                resourceImpl.setMediaType(result.getString("MEDIA_TYPE"));
                resourceImpl.setAuthorUserName(result.getString("CREATOR"));
                resourceImpl.setCreatedTime(result.getTimestamp("CREATED_TIME"));
                resourceImpl.setLastUpdaterUserName(result.getString("LAST_UPDATOR"));
                resourceImpl.setLastModified(result.getTimestamp("LAST_UPDATED_TIME"));
                resourceImpl.setDescription(result.getString("DESCRIPTION"));
                resourceImpl.setContentID(result.getString("CONTENT_ID"));
            }

            ps.close();

            return resourceImpl;

        } catch (SQLException e) {

            String msg = "Failed to get resource " + resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public String addContentVersion(InputStream contentStream) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String contentID = null;

        if (contentStream != null) {

            // we should first write the content stream to a temporary file because JDBC
            // needs the length of the binary stream.

            File tempFile;
            try {
                tempFile = File.createTempFile("reg", "tmp");
                OutputStream fileStream = new BufferedOutputStream(new FileOutputStream(tempFile));
                InputStream bufferedIn = new BufferedInputStream(contentStream);

                byte[] dataChunk = new byte[1024];
                int byteCount;
                while ((byteCount = bufferedIn.read(dataChunk)) != -1) {
                    fileStream.write(dataChunk, 0, byteCount);
                }
                fileStream.flush();
                bufferedIn.close();
                fileStream.close();

            } catch (IOException e) {

                String msg = "Failed to write the resource content to a temporary file, " +
                        "before writing to the database. " + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);
            }

            try {

                contentID = UUID.randomUUID().toString();

                InputStream tempFileStream = new BufferedInputStream(new FileInputStream(tempFile));
                long contentLength = tempFile.length();

                String sql = "INSERT INTO CONTENT_VERSION " +
                        "(CONTENT_VERSION_ID, CONTENT_DATA) VALUES (?, ?)";

                PreparedStatement ps = conn.prepareStatement(sql);
                ps.setString(1, contentID);
                ps.setBinaryStream(2, tempFileStream, (int)contentLength);
                ps.executeUpdate();
                ps.close();

            } catch (FileNotFoundException e) {
                String msg = "Failed to read resource content from the temporary file. " +
                        e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);

            } catch (SQLException e) {

                String msg = "Failed to write resource content to the database. " + e.getMessage();
                log.error(msg, e);
                throw new RegistryException(msg, e);
            }
        }

        return contentID;
    }

    public InputStream getResourceContentStream(String contentID, Connection conn)
            throws RegistryException {

        InputStream contentStream = null;

        try {
            String sql = "SELECT CONTENT_DATA FROM CONTENT_VERSION WHERE CONTENT_VERSION_ID=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, contentID);

            ResultSet result = ps.executeQuery();
            if (result.next()) {
                contentStream = getDisconnectedStream(result.getBinaryStream("CONTENT_DATA"));
            }
            ps.close();

        } catch (SQLException e) {
            String msg = "Faild to get resource content from the database. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);

        } catch (IOException e) {
            String msg = "Failed to write resource content to a temporary file. " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return contentStream;
    }

    private InputStream getDisconnectedStream(InputStream connectedStream) throws IOException {

        if (connectedStream == null) {
            return null;
        }

        File tempFile = File.createTempFile("reg", "tmp");
        tempFile.deleteOnExit();

        BufferedOutputStream out = new BufferedOutputStream(new FileOutputStream(tempFile));
        byte[] contentChunk = new byte[1024];
        int byteCount;
        while ((byteCount = connectedStream.read(contentChunk)) != -1) {
            out.write(contentChunk, 0, byteCount);
        }
        out.flush();
        out.close();
        connectedStream.close();

        return new BufferedInputStream(new FileInputStream(tempFile));
    }

    public ResourceDO getResourceVersionDO(String resourceID, long snapshotID)
            throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT R.RID, R.VERSION, R.PATH, R.MEDIA_TYPE, R.COLLECTION, " +
                "R.CREATOR, R.CREATED_TIME, R.LAST_UPDATOR, R.LAST_UPDATED_TIME, " +
                "R.DESCRIPTION, R.CONTENT_ID, R.ASSOCIATED_SNAPSHOT_ID " +
                "FROM RESOURCE_VERSION R WHERE R.RID=? AND R.VERSION=" +
                "(SELECT S.VERSION FROM SNAPSHOT_RESOURCE_VERSION S " +
                "WHERE S.SNAPSHOT_ID=? AND S.RID=?)";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, snapshotID);
            ps.setString(3, resourceID);

            ResourceDO resourceVersionDO = null;
            ResultSet result = ps.executeQuery();
            if (result.next()) {
                resourceVersionDO = new ResourceDO();

                resourceVersionDO.setID(result.getString("RID"));
                resourceVersionDO.setVersion(result.getLong("VERSION"));
                resourceVersionDO.setPath(result.getString("PATH"));
                resourceVersionDO.setMediaType(result.getString("MEDIA_TYPE"));
                resourceVersionDO.setCollection(result.getInt("COLLECTION"));
                resourceVersionDO.setAuthor(result.getString("CREATOR"));
                resourceVersionDO.setCreatedOn(result.getTimestamp("CREATED_TIME"));
                resourceVersionDO.setLastUpdator(result.getString("LAST_UPDATOR"));
                resourceVersionDO.setLastUpdatedOn(result.getTimestamp("LAST_UPDATED_TIME"));
                resourceVersionDO.setDescription(result.getString("DESCRIPTION"));
                resourceVersionDO.setContentID(result.getString("CONTENT_ID"));
                resourceVersionDO.setAssociatedSnapshotID(result.getLong("ASSOCIATED_SNAPSHOT_ID"));
            }

            return resourceVersionDO;

        } catch (SQLException e) {

            String msg = "Failed to read resource version data for resource " +
                    resourceID + " for snapshot ID " + snapshotID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }
    }

    public List<String> getChildIDs(String resourceID, long version) throws RegistryException {

        List <String> childIDs = new ArrayList();

        Connection conn = Transaction.getConnection();

        try {
            String sql = "SELECT CHILD_RID FROM DEPENDENCY_VERSION " +
                    "WHERE PARENT_RID=? AND PARENT_VERSION=?";

            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            ResultSet results = ps.executeQuery();
            while (results.next()) {
                childIDs.add(results.getString("CHILD_RID"));
            }
            ps.close();

        } catch (SQLException e) {

            String msg = "Failed to get child resource IDs of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

        return childIDs;
    }

    /**
     * Given a resource ID and a version number there may or may not be a snapshot ID. That is,
     * whenever a snapshot is created, an associating version of the resource is prepared. But,
     * when creating a snapshot of a collection, new versions will be created for all dirty child
     * resources/collections. Such resources/collection versions do not have a matching snapshot ID.
     * Such versions are just a part of a snapshot of another collection. And it is possible to
     * have more than one matching snapshot for a given resource ID and a version. A simple
     * example is creating two snapshots of a resource, without modifying the resource.
     *
     * This method returns the latest (highest) snapshot ID for the given resource ID and
     * the version. If a mathcing snapshot is not found, it will return -1.
     *
     * @param resourceID UUID of the resource
     * @param version Version number of the resource.
     * @return Matching snapshot number if available. -1 otherwise.
     * @throws RegistryException
     */
    public long getMatchingSnapshotID(String resourceID, long version) throws RegistryException {

        List <Long> actualSnapshotIDs = getSnapshotIDsForResourceID(resourceID);
        if (actualSnapshotIDs.size() == 0) {
            return -1;
        }

        List <Long> possibleSnapshotIDs = getSnapshotIDs(resourceID, version);

        long matchingSID = -1;
        for (int i = 0; i < actualSnapshotIDs.size(); i++) {
            long actualSID = actualSnapshotIDs.get(i);
            if (possibleSnapshotIDs.contains(actualSID)) {
                matchingSID = actualSID;
            }
        }

        return matchingSID;
    }

    private List<Long> getSnapshotIDsForResourceID(String resourceID) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT SNAPSHOT_ID FROM SNAPSHOT WHERE ROOT_ID=? ORDER BY SNAPSHOT_ID DESC";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);

            ResultSet results = ps.executeQuery();
            List <Long> snapshotIDs = new ArrayList <Long> ();
            while(results.next()) {
                long snapshotNumber = results.getLong("SNAPSHOT_ID");
                snapshotIDs.add(snapshotNumber);
            }
            ps.close();

            return snapshotIDs;

        } catch (SQLException e) {

            String msg = "Failed to get snapshot numbers of resource " +
                    resourceID + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }

    private List<Long> getSnapshotIDs(String resourceID, long version) throws RegistryException {

        Connection conn = Transaction.getConnection();

        String sql = "SELECT SNAPSHOT_ID FROM SNAPSHOT_RESOURCE_VERSION WHERE RID=? AND VERSION=?";

        try {
            PreparedStatement ps = conn.prepareStatement(sql);
            ps.setString(1, resourceID);
            ps.setLong(2, version);

            ResultSet results = ps.executeQuery();

            List <Long> sids = new ArrayList <Long> ();
            while (results.next()) {
                sids.add(results.getLong("SNAPSHOT_ID"));
            }
            ps.close();

            return sids;

        } catch (SQLException e) {

            String msg = "Failed to matching snapshot IDs for resource ID " +
                    resourceID + " and version " + version + ". " + e.getMessage();
            log.error(msg, e);
            throw new RegistryException(msg, e);
        }

    }
}
