/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.handlers.builtin;

import org.wso2.registry.jdbc.handlers.Handler;
import org.wso2.registry.jdbc.handlers.RequestContext;
import org.wso2.registry.jdbc.Repository;
import org.wso2.registry.jdbc.dao.CommentsDAO;
import org.wso2.registry.*;
import org.wso2.registry.session.CurrentSession;
import org.wso2.registry.secure.AuthorizationFailedException;
import org.wso2.registry.utils.AuthorizationUtil;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.users.UserStoreException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;

public class CommentURLHandler extends Handler {

    private static final Log log = LogFactory.getLog(CommentURLHandler.class);

    private CommentsDAO commentsDAO = new CommentsDAO();

    public CommentURLHandler() {}

    public CommentURLHandler(Repository repository, Registry registry) {
        super(repository, registry);
    }

    public Resource get(RequestContext requestContext) throws RegistryException {

        String url = requestContext.getResourcePath();
        String[] parts = url.split(RegistryConstants.URL_SEPARATOR);

        if (parts.length != 2) {
            return null;
        }

        if (parts[0].length() == 0) {
            return null;
        }

        if (!parts[1].startsWith("comments:")) {
            return null;
        }

        String commentID = parts[1].substring("comments:".length());
        if (!(commentID.length() > 0)) {
            return null;
        }

        long cID;
        try {
            cID = Long.parseLong(commentID);

        } catch (NumberFormatException e) {
            // note that this might not be an exceptional scenario. there could be a different URL
            // form, which contains strings after "comment".
            // it is just that it is not the URL we expect here
            return null;
        }

        Comment comment = commentsDAO.getComment(cID);
        String resourcePath = commentsDAO.getResourcePathOfComment(cID);

        if (comment == null) {
            String msg = "Requested comment with ID: " + cID + " is not available.";
            log.info(msg);
            throw new RegistryException(msg);
        }

        requestContext.setProcessingComplete(true);
        return comment;
    }

    public void put(RequestContext requestContext) throws RegistryException {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    public void importResource(RequestContext requestContext) throws RegistryException {

    }

    public void delete(RequestContext requestContext) throws RegistryException {

        requestContext.setProcessingComplete(false);
        String path = requestContext.getResourcePath();
        String[] parts = path.split(RegistryConstants.URL_SEPARATOR);

        if (parts.length != 2) {
            return;
        }

        if (parts[0].length() == 0) {
            return;
        }

        if (!parts[1].startsWith("comments:")) {
            return;
        }

        String commentID = parts[1].substring("comments:".length());
        if (!(commentID.length() > 0)) {
            return;
        }

        long cID;
        try {
            cID = Long.parseLong(commentID);

        } catch (NumberFormatException e) {
            // note that this might not be an exceptional scenario. there could be a different URL
            // form, which contains strings after "comment".
            // it is just that it is not the URL we expect here
            return;
        }

        String userID = CurrentSession.getUser();
        String authPath = AuthorizationUtil.getAuthorizationPath(path);
        String commentAuthor;

        Comment comment = commentsDAO.getComment(cID);
        commentAuthor = comment.getUser();

        // check if the current user has permission to delete this comment.
        // users who have PUT permission on the commented resource can delete any comment on
        // that resource. Any user can delete his own comment.

        try {
            UserRealm realm = CurrentSession.getRealm();

            if (!userID.equals(commentAuthor) &&
                    !realm.getAuthorizer().isUserAuthorized(userID, authPath, ActionConstants.PUT)) {

                String msg = "User: " + userID +
                        " is not authorized to delete the comment on the resource: " +
                        authPath;
                log.info(msg);
                throw new AuthorizationFailedException(msg);
            }

        } catch (UserStoreException e) {
            //
        }

        commentsDAO.deleteComment(cID);

        requestContext.setProcessingComplete(true);
    }

    public void putChild(RequestContext requestContext) throws RegistryException {
        //To change body of implemented methods use File | Settings | File Templates.
    }

    public void importChild(RequestContext requestContext) throws RegistryException {

    }
}
