/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.handlers.filters;

import org.wso2.registry.jdbc.handlers.RequestContext;
import org.wso2.registry.Resource;
import org.wso2.registry.RegistryException;
import org.wso2.registry.Collection;
import org.wso2.registry.utils.RegistryUtils;

public class MediaTypeMatcher extends Filter {

    private String mediaType;

    public MediaTypeMatcher() {}

    public MediaTypeMatcher(String mediaType) {
        this.mediaType = mediaType;
    }

    /**
     * Matches if the media type of the current resource is equal to the handler's media type.
     *
     * @param requestContext RequestContext for the current request
     * @return true if the media type of the current resource is equal to the handler's media type.
     * @throws RegistryException
     */
    public boolean handleGet(RequestContext requestContext) throws RegistryException {

        Resource resource = requestContext.getResource();
        if (resource == null) {
            resource = repository.get(requestContext.getResourcePath());
            requestContext.setResource(resource);
        }

        if (resource != null) {
            String mType = resource.getMediaType();
            if (mType != null && mType.equals(mediaType)) {
                return true;
            }
        }

        return false;
    }

    public boolean handlePut(RequestContext requestContext) throws RegistryException {

        Resource resource = requestContext.getResource();
        if (resource == null) {
            return false;
        }

        String mType = resource.getMediaType();
        return mType != null && mType.equals(mediaType);

    }

    public boolean handleImportResource(RequestContext requestContext) throws RegistryException {

        Resource metadataResource = requestContext.getResource();
        if (metadataResource == null) {
            return false;
        }

        String mType = metadataResource.getMediaType();
        return mType != null && mType.equals(mediaType);

    }

    public boolean handleDelete(RequestContext requestContext) throws RegistryException {

        Resource resource = requestContext.getResource();
        if (resource == null) {
            resource = repository.get(requestContext.getResourcePath());
            requestContext.setResource(resource);
        }

        if (resource != null) {
            String mType = resource.getMediaType();
            if (mType != null && mType.equals(mediaType)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Matches if the media type of the parent collection of the current resource is equal to
     * the handler's media type.
     *
     * @param requestContext RequestContext for the current request
     * @return true if the media type of the parent collection of the current resource is equal
     * to the handler's media type.
     * 
     * @throws RegistryException
     */
    public boolean handlePutChild(RequestContext requestContext) throws RegistryException {

        Collection parentCollection = requestContext.getParentCollection();
        if (parentCollection == null) {
            String parentPath = requestContext.getParentPath();
            if (parentPath == null) {
                parentPath = RegistryUtils.getParentPath(requestContext.getResourcePath());
                requestContext.setParentPath(parentPath);
            }

            parentCollection = (Collection) repository.get(parentPath);
            requestContext.setParentCollection(parentCollection);
        }

        if (parentCollection != null) {
            String parentMediaType = parentCollection.getMediaType();
            if (parentMediaType != null && parentMediaType.equals(mediaType)) {
                return true;
            }
        }

        return false;
    }

    public boolean handleImportChild(RequestContext requestContext) throws RegistryException {

        Collection parentCollection = requestContext.getParentCollection();
        if (parentCollection == null) {
            String parentPath = requestContext.getParentPath();
            if (parentPath == null) {
                parentPath = RegistryUtils.getParentPath(requestContext.getResourcePath());
                requestContext.setParentPath(parentPath);
            }

            parentCollection = (Collection) repository.get(parentPath);
            requestContext.setParentCollection(parentCollection);
        }

        if (parentCollection != null) {
            String parentMediaType = parentCollection.getMediaType();
            if (parentMediaType != null && parentMediaType.equals(mediaType)) {
                return true;
            }
        }

        return false;
    }

    public String getMediaType() {
        return mediaType;
    }

    public void setMediaType(String mediaType) {
        this.mediaType = mediaType;
    }
}
